;; Generated package description from so-long.el  -*- no-byte-compile: t -*-
(define-package "so-long" "1.1.2" "Say farewell to performance problems with minified code." '((emacs "24.4")) :commit "045a4fe94c18cd36ef297e62a80cdff449af3aa5" :authors '(("Phil Sainty" . "psainty@orcon.net.nz")) :maintainer '("Phil Sainty" . "psainty@orcon.net.nz") :keywords '("convenience") :url "https://savannah.nongnu.org/projects/so-long")
